/*
** rts/print.c
**
** Viper RTS Print (of data structures) functions
**
** Copyright (c) 1998 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>

#include "rts/rts.h"
#include "rts/print.h"

const char *
rts_i2yesno(int i)
{
    switch (i)
    {
      case 0: return "No";
      case 1: return "Yes";
      default: return "Unknown";
    }
}


char *
rts_i2bs(unsigned int iv)
{
    static char buf[256];
    char *cp;

    cp = buf;
    while (iv)
    {
	*cp++ = (iv&1)+'0';
	iv>>=1;
    }

    if (cp == buf)
	*cp++ = '0';

    *cp++ = 'b';
    *cp = '\0';

    return buf;
}
    
static void
rts_print_channel_status(FILE *fp,
			 int version,
			 unsigned int s)
{
    if ((s & 0x01) == 1)
    {
	fprintf(fp, "Slot not available");
	return;
    }

    if ((s & 0x02) == 0)
    {
	fprintf(fp, "Slot is empty");
	return;
    }

    
   
    fprintf(fp, "Selftest %s",
	    ((s & 0x04) == 0) ? "OK" : "Failed");

    fprintf(fp, ", %s", ((s & 0x08) == 0) ? "Narrow" : "Wide");
    fprintf(fp, ", %s", ((s & 0x10) == 0) ? "SE" : "HVD");

    if (version <= 4)
	fprintf(fp, ", %s", ((s & 0x20) == 0) ? "Non-PCI" : "PCI");
    else
    {
	if ((s & 0x20) == 1)
	    fprintf(fp, ", FC-Terminators");
    }
    
    fprintf(fp, ", %s", ((s & 0x40) == 0) ? "Host" : "Disk");
    
    if ((s & 0x80) == 1)
	fprintf(fp, ", FAST20");
		
	    
}
    
void
rts_print_hex(FILE *fp,
	      Byte *bp,
	      int len)
{
    while (len-- > 0)
    {
	fprintf(fp, "%02X", *bp++);
	if (len > 0)
	    putc(' ', fp);
    }
}



void
rts_print_host_parameters(FILE *fp,
			  int version,
			  struct rts_host_parameters *hp)
{
    int i;

    
    fprintf(fp, "%-35s%d\n", "Channel", hp->channel);
    fprintf(fp, "%-35s%d\n", "SCSI ID", hp->scsi_id);
    fprintf(fp, "%-35s%s\n", "Tag Queuing", hp->tag_queuing ? "On" : "Off");
    fprintf(fp, "%-35s%s\n", "Sync Mode", hp->sync_mode ? "On" : "Off");
    fprintf(fp, "%-35s%d\n", "Sync Rate", hp->sync_rate);
    fprintf(fp, "%-35s%d\n", "Bus Width", hp->bus_width ? 16 : 8);
    fprintf(fp, "%-35s", "Termination");
    switch (hp->termination)
    {
      case 0:
	fputs("None\n", fp);
	break;
      case 1:
	fputs("Partial\n", fp);
	break;
      case 2:
	fputs("Full\n", fp);
	break;
      default:
	fputs("Unknown\n", fp);
    }

    if (version >= 9)
     fprintf(fp, "%-35s%02X\n", "Inquiry Peripheral",
	     hp->inquiry_peripheral);

    fprintf(fp, "LUN/Redundancy Group mapping:\n");

    for (i = 0; i < 32 ; i++)
	if (hp->host_lun_map[i] != 255 && i != hp->host_lun_map[i])
	    break;

    if (i == 32)
	fprintf(fp, "\tAll LUNs (32) map 1-1 to Redundancy Groups\n");
    else
    {
	for (i = 0; i < 32; i++)
	    if (hp->host_lun_map[i] != 255)
		fprintf(fp, "\tLUN %2d -> Group %2d\n",
			i, hp->host_lun_map[i]);
    }
}


void
rts_print_redundancy_group_information(FILE *fp,
				       int version,
				       struct rts_redundancy_group_information *rp)
{
    time_t bt;
    int i;
    
    
    fprintf(fp, "%-35s%d\n", "Redundancy group:", rp->redundancy_group);
    fprintf(fp, "%-35s%d\n", "RAID set:", rp->raid_set);
    fprintf(fp, "%-35s%s\n", "RAID status:", rts_status2str(rp->raid_status));
    if (rp->raid_status == 0)
	return;

    fprintf(fp, "%-35s%s\n", "RAID level:", rts_raid2str(rp->raid_level));
    
    if (rp->host_0_lun != 255)
	fprintf(fp, "%-35s%d\n", "Host channel 0 LUN:", rp->host_0_lun);
    if (rp->host_1_lun != 255)
	fprintf(fp, "%-35s%d\n", "Host channel 1 LUN:", rp->host_1_lun);
    if (rp->host_2_lun != 255)
	fprintf(fp, "%-35s%d\n", "Host channel 2 LUN:", rp->host_2_lun);
    if (rp->host_3_lun != 255)
	fprintf(fp, "%-35s%d\n", "Host channel 3 LUN:", rp->host_3_lun);

    bt = rts_mktime(&rp->time_stamp);
    fprintf(fp, "%-35s%s", "Created:", ctime(&bt));
    fprintf(fp, "%-35s%d blocks\n", "Chunk size:", rts_w2i(rp->chunk_size));
    fprintf(fp, "%-35s%d blocks (%d GB)\n", "Capacity:",
	    rts_q2i(rp->capacity),
	    rts_q2i(rp->capacity)/2/1024/1024);
    
    fprintf(fp, "%-35s%d\n", "Members:", rp->number_of_members);
    for (i = 0; i < rp->number_of_members; i++)
    {
	fprintf(fp, "Member %d:                          Channel %d, SCSI ID %d\n",
		i,
		rp->member_descriptor[i].channel,
		rp->member_descriptor[i].scsi_id);
    }
}

const char *
rts_i2model(int i)
{
    switch (i)
    {
      case 0: return "5300";
      case 1: return "5400";
      case 2: return "5500";
      case 3: return "5600";
      case 4: return "6560";
      case 5: return "5620";
      case 6: return "5640";
      case 7: return "5330";
      case 8: return "5440";
      default: return "Unknown";
    }
}

const char *
rts_i2proctype(int i)
{
    switch (i)
    {
      case 0: return "33300";
      case 1: return "33310";
      default: return "Unknown";
    }
}

void
rts_print_static_system_information(FILE *fp,
				    int version,
				    struct rts_static_system_information *sp)
{
    int i;
    
    fprintf(fp, "%-35s%.9s\n",
	    "Firmware version:",
	    sp->firmware_version);
    fprintf(fp, "%-35s%d\n",
	    "RTS version:",
	    sp->rts_version);
    fprintf(fp, "%-35s%s\n",
	    "Model number:",
	    rts_i2model(sp->model_number));
    fprintf(fp, "%-35s%s\n",
	    "Processor type:",
	    rts_i2proctype(sp->processor_type));
    fprintf(fp, "%-35s%d MHz\n",
	    "Processor clock:",
	    sp->processor_clock);
    fprintf(fp, "%-35s%s\n",
	    "Processor memory:",
	    sp->processor_memory ? "Cache DRAM" : "Processor DRAM");
    fprintf(fp, "%-35s%d MB\n",
	    "Cache size:",
	    rts_w2i(sp->cache_size));
    fprintf(fp, "%-35s%d MB\n",
	    "SIMM 0A size:",
	    sp->simm_0a_size);
    fprintf(fp, "%-35s%d MB\n",
	    "SIMM 0B size:",
	    sp->simm_0b_size);
    fprintf(fp, "%-35s%d MB\n",
	    "SIMM 1A size:",
	    sp->simm_1a_size);
    fprintf(fp, "%-35s%d MB\n",
	    "SIMM 1B size:",
	    sp->simm_1b_size);
    fprintf(fp, "%-35s%.9s\n",
	    "Serial number:",
	    sp->serial_number);
    fprintf(fp, "%-35s%d/%d/%d (MM/DD/YY)\n",
	    "Manufacture date:",
	    sp->manufacture_month,
	    sp->manufacture_date,
	    sp->manufacture_year);
    
    for (i = 0; i < 10; i++)
    {
	fprintf(fp, "Channel %d Status:                  ", i);
	rts_print_channel_status(fp, version, sp->channel_status[i]);
	putc('\n', fp);
    }
    
    fprintf(fp, "%-35s%d\n",
	    "DRAM 0 vendor ID:",
	    sp->dram_0_vendor_id);
    fprintf(fp, "%-35s%d\n",
	    "DRAM 1 vendor ID:",
	    sp->dram_1_vendor_id);
    fprintf(fp, "%-35s%d\n",
	    "XOR vendor ID:",
	    sp->xor_vendor_id);
    fprintf(fp, "%-35s%d\n",
	    "PIC vendor ID:",
	    sp->pic_vendor_id);
}

static void
rts_print_bm_id(FILE *fp,
		unsigned int bm)
{
    int id = 0;

    if (bm == 0)
    {
	fprintf(fp, "None");
	return;
    }
    
    while (bm)
    {
	if (bm&1)
	    fprintf(fp, "%d", id);
	
	bm>>=1;
	if (bm&1)
	    fprintf(fp, ", ");

	++id;
    }
}

void
rts_print_dynamic_system_information(FILE *fp,
				     int version,
				     struct rts_dynamic_system_information *sp)
{
    fprintf(fp, "%-35s", "Online redundancy groups:");
    rts_print_bm_id(fp, rts_q2i(sp->online_redundancy_groups));;
    
    fprintf(fp, "\n%-35s", "Online RAID sets:");
    rts_print_bm_id(fp, rts_q2i(sp->online_raid_sets));
	    
    fprintf(fp, "\n%-35s", "Creating RAID sets:");
    rts_print_bm_id(fp, rts_q2i(sp->creating_raid_sets));

    fprintf(fp, "\n%-35s", "Channel 0 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_0_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 1 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_1_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 2 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_2_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 3 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_3_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 4 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_4_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 5 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_5_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 6 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_6_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 7 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_7_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 8 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_8_disk_drives));
    
    fprintf(fp, "\n%-35s", "Channel 9 disk drives:");
    rts_print_bm_id(fp, rts_w2i(sp->channel_9_disk_drives));

    fprintf(fp, "\n%-35s", "Remote access level:");
    switch (sp->remote_access_level)
    {
      case 0:
	puts("Limited Access");
	break;

      case 1:
	puts("Vendor Access");
	break;

      case 2:
	puts("User Access");
	break;

      default:
	printf("Unknown (#%d)\n", sp->remote_access_level);
    }
}


void
rts_print_system_parameters(FILE *fp,
			    int version,
			    struct rts_system_parameters *sp)
{
    fprintf(fp, "%-35s%s\n", "Password checking:",
	    rts_i2yesno(sp->password_checking));
    fprintf(fp, "%-35s%d\n", "Rebuild rate:",
	    sp->rebuild_rate);
    fprintf(fp, "%-35s%d\n", "Create rate:",
	    sp->create_rate);
    fprintf(fp, "%-35s%s\n", "UPS installed:",
	    rts_i2yesno(sp->ups_installed));
    fprintf(fp, "%-35s%s\n", "Channel 1 module type:",
	    sp->ch1_module_type ? "Disk" : "Host");
    fprintf(fp, "%-35s%s\n", "Channel 2 module type:",
	    sp->ch2_module_type ? "Disk" : "Host");
    fprintf(fp, "%-35s%s\n", "Channel 3 module type:",
	    sp->ch3_module_type ? "Disk" : "Host");
    
    if (version < 2)
	return;
    
    fprintf(fp, "%-35s%s\n", "Read ahead enabled:",
	    rts_i2yesno(sp->read_ahead_enabled));
    fprintf(fp, "%-35s%d blocks\n", "Read ahead limit:",
	    sp->read_ahead_limit * 32);
    
    if (version < 3)
	return;
    
    fprintf(fp, "%-35s%d s\n", "Validation delay:",
	    sp->validation_delay);
    fprintf(fp, "%-35s%s\n", "Twelve volt sensor enabled:",
	    rts_i2yesno(sp->twelve_volt_sensor_enabled));

    if (version < 6)
	return;
    
    fprintf(fp, "%-35s%s\n", "Host WCE enabled:",
	    rts_i2yesno(sp->host_wce_enabled));
    fprintf(fp, "%-35s%d s\n", "Drive command timeout:",
	    sp->drive_command_timeout);
    
    if (version < 9)
	return;
    
    fprintf(fp, "%-35s%s\n", "Reset propagation:",
	    rts_i2yesno(sp->reset_propagation));
    fprintf(fp, "%-35s%s\n", "Environmental alerts:",
	    rts_i2yesno(sp->environmental_alerts));
    fprintf(fp, "%-35s%d s\n", "Spinup staggering:",
	    sp->spinup_staggering);
}


static double
f2c(double f)
{
    return (f - 32) / 1.8;
}

static int
tl2f(int tl)
{
    if (tl == 0)
	return 0;

    return tl*5+95;
}


void
rts_print_vendor_parameters(FILE *fp,
			    int version,
			    struct rts_vendor_parameters *vp)
{
    fprintf(fp, "%-35s%.16s\n",
	    "Display Vendor:",
	    vp->system_display_vendor);
    
    fprintf(fp, "%-35s%.16s\n",
	    "Display Model:",
	    vp->system_display_model);
    
    fprintf(fp, "%-35s%.8s\n",
	    "Inquiry Vendor:",
	    vp->inquiry_response_vendor);
    
    fprintf(fp, "%-35s%.16s\n",
	    "Inquiry Model:",
	    vp->inquiry_response_model);
    
    
    if (version >= 2)
    {
	fprintf(fp, "%-35s%d %%\n",
		"Write back cache:",
		vp->percent_write_back_cache);
    }
    
    if (version >= 5)
    {
	fprintf(fp, "%-35s%d F (%d C)\n",
		"Enclosure severe temperature:",
		tl2f(vp->enclosure_severe_level_temp),
		(int) (f2c(tl2f(vp->enclosure_severe_level_temp))+0.5));
	
	fprintf(fp, "%-35s%d F (%d C)\n",
		"Enclosure warning temperature:",
		tl2f(vp->enclosure_warning_level_temp),
		(int) (f2c(tl2f(vp->enclosure_warning_level_temp))+0.5));
    }
    
    if (version >= 7)
    {
	fprintf(fp, "%-35s%s\n",
		"Extended acknowledge mode:",
		rts_i2yesno(vp->extended_acknowledge_mode));
    }
}

static const char *
rts_i2envstatus(int status)
{
    switch (status)
    {
      case 0: return "Not installed";
      case 1: return "Normal";
      case 2: return "Below normal";
      case 3: return "Above normal";
      case 4: return "Far below normal";
      case 5: return "Far above normal";
      case 6: return "Back to normal";
      default: return "Unknown";
    }
}



void
rts_print_environmental_information(FILE *fp,
				    int version,
				    struct rts_environmental_information *ep)
{
    fprintf(fp, "%-35s%.2f V - %s\n",
	    "System board +5V level:",
	   ((double) rts_w2i(ep->five_volt_level)) / 100.0,
	   rts_i2envstatus(ep->five_volt_status));
    
    fprintf(fp, "%-35s%.2f V - %s\n",
	    "System board +12V level:",
	   ((double) rts_w2i(ep->twelve_volt_level)) / 100.0,
	   rts_i2envstatus(ep->twelve_volt_status));
    
    fprintf(fp, "%-35s%.2f V - %s\n",
	    "SCSI Termination +5V level:",
	   ((double) rts_w2i(ep->scsi_termination_volt_level)) / 100.0,
	   rts_i2envstatus(ep->scsi_termination_volt_status));
    
    fprintf(fp, "%-35s%.2f V - %s\n",
	    "DRAM bank 0 +5V level:",
	   ((double) rts_w2i(ep->dram_0_volt_level)) / 100.0,
	   rts_i2envstatus(ep->dram_0_volt_status));
    
    fprintf(fp, "%-35s%.2f V - %s\n",
	    "DRAM bank 1 +5V level:",
	   ((double) rts_w2i(ep->dram_1_volt_level)) / 100.0,
	   rts_i2envstatus(ep->dram_1_volt_status));
    
    fprintf(fp, "%-35s%.2f V - %s\n",
	    "Backup battery volt level:",
	   ((double) rts_w2i(ep->backup_battery_volt_level)) / 100.0,
	   rts_i2envstatus(ep->backup_battery_volt_status));
    
    fprintf(fp, "%-35s%d F (%d C) - %s\n",
	    "Board temperature level:",
	   ep->board_temperature_level,
	   (int) (f2c(ep->board_temperature_level)+0.5),
	   rts_i2envstatus(ep->board_temperature_status));
    
    fprintf(fp, "%-35s%d F (%d C) - %s\n",
	    "Enclosure temperature level:",
	   ep->enclosure_temperature_level,
	   (int) (f2c(ep->enclosure_temperature_level)+0.5),
	   rts_i2envstatus(ep->enclosure_temperature_status));
    
    fprintf(fp, "%-35s%s\n", "Calibration Status:",
	   ep->calibration_status ? "Calibrated" : "Not calibrated");
    
    if (version < 3)
	return;

    fprintf(fp, "%-35s", "UPS Outputs:");
    if (ep->ups_outputs == 0)
	fprintf(fp, "Normal");
    
    if (ep->ups_outputs & 0x01)
	fprintf(fp, "AC Power Lost");
    if (ep->ups_outputs & 0x02)
    {
	if (ep->ups_outputs & 0x01)
	    fprintf(fp, ", ");
	fprintf(fp, "Two Minute Warning");
    }
    putchar('\n');
    
    fprintf(fp, "%-35s", "UPS Status:");
    switch (ep->ups_status)
    {
      case 0:
	puts("Not installed");
	break;
	
      case 1:
	puts("Normal");
	break;
	
      case 6:
	puts("Back to normal");
	break;
	
      case 7:
	puts("AC Power Lost");
	break;
	
      case 8:
	puts("Two Minute Warning");
	break;
	
      default:
	fprintf(fp, "Unknown Status: #%d\n", ep->ups_status);
    }
}
