/*
** viper.c
**
** Get(/set) status/parameters for a CMD ViperII RAID controller
**
** Copyright (c) 1998-2002 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <math.h>
#include <signal.h>

#include <rts/rts.h>

#include "daemon.h"

extern char version[];

static char *argv0 = "viper";
static int debug = 0;

static int check_netsaint = 0;

int temp_warning = 0; /* C */
int temp_severe  = 0; /* C */

static int readonly = 0;

static int print_vendor = 0;
static int print_env = 0;
static int print_static = 0;
static int print_dynamic = 0;
static int print_system = 0;
static int print_host = 0;
static int print_redgrp = 0;
static int print_misc = 0;


char **devtab = NULL;
int devidx = 0;
int devtabsize = 0;

static double
f2c(double f)
{
    return (f - 32) / 1.8;
}

static int
tl2f(int tl)
{
    if (tl == 0)
	return 0;

    return tl*5+95;
}

void
do_time(RTS *rp)
{
    int code;

    time_t c_t, o_t;

    code = rts_read_date_and_time(rp, &c_t);
    if (code)
    {
	perror("mktime");
	rts_perror(stderr, "rts_read_date_and_time", rp);
	exit(1);
    }
    
    time(&o_t);
    
    if (debug)
    {
	fprintf(stderr, "Our time:   %s\n", ctime(&o_t));
	fprintf(stderr, "Viper time: %s\n", ctime(&c_t));
    }
    
    if (fabs(difftime(o_t, c_t)) > 5)
    {
	
	code = rts_begin_session(rp, 0);
	if (code)
	{
	    rts_perror(stderr, "rts_begin_session", rp);
	    exit(1);
	}
	
	code = rts_write_date_and_time(rp, o_t);
	if (code)
	{
	    rts_perror(stderr, "rts_write_date_and_time", rp);
	    rts_end_session(rp);
	    exit(1);
	}

	
	code = rts_end_session(rp);
	if (code)
	{
	    rts_perror(stderr, "rts_end_session", rp);
	    exit(1);
	}
	
	syslog(LOG_NOTICE, "Time set on Viper");
    }
}


void
usage(FILE *fp)
{
    fprintf(fp, "Usage: %s [options] <device>\n", argv0);
    fprintf(fp, "Options:\n");
    fprintf(fp, "\t-V          Print version\n");
    fprintf(fp, "\t-d          Debug mode\n");
    fprintf(fp, "\t-h          Display this information\n");
}

static int
check_vt_status(int status)
{
    switch (status)
    {
      case 0:
      case 1:
      case 6:
	return 0;

      case 2:
      case 3:
	return 1;

      case 4:
      case 5:
	return 2;

      default:
	return -1;
    }
}


int
check_temp(RTS *rp,
	   char **msg,
	   struct rts_vendor_parameters *vp,
	   struct rts_environmental_information *eip)
{
    static char msgbuf[256], *cp;
    int ecode = 0, necode = 0;
    int brd_temp_current = 0;
    int enc_temp_current = 0;
    

    
#if RTS_VERSION >= 5
    if (temp_warning == 0)
	temp_warning = f2c(tl2f(vp->enclosure_warning_level_temp))+0.5;
#endif

#if RTS_VERSION >= 2
    if (temp_severe == 0)
	temp_severe = f2c(tl2f(vp->enclosure_severe_level_temp))+0.5;
#endif

    if (eip->enclosure_temperature_level > 0)
	enc_temp_current = f2c(eip->enclosure_temperature_level)+0.5;
    else
	enc_temp_current = -1;
    
    brd_temp_current = f2c(eip->board_temperature_level)+0.5;

    
    if (enc_temp_current > 0)
    {
	if (temp_warning > 0 && enc_temp_current >= temp_warning && ecode < 1)
	    ecode = 1;
	
	if (temp_severe > 0 && enc_temp_current >= temp_severe && ecode < 2)
	    ecode = 2;
    }

    necode = check_vt_status(eip->board_temperature_status);
    if (necode > ecode)
	ecode = necode;
	
    necode = check_vt_status(eip->enclosure_temperature_status);
    if (necode > ecode)
	ecode = necode;
	
    cp = msgbuf;
    sprintf(cp, "Temp: %dC", brd_temp_current);
    while (*cp) ++cp;
    if (enc_temp_current > 0)
    {
	sprintf(cp, ", Enc=%dC", enc_temp_current);
	while (*cp) ++cp;
    }

    if (temp_severe > 0 || temp_warning > 0)
    {
	if (temp_severe > 0 && enc_temp_current >= temp_severe)
	    sprintf(cp, " - over Severe Limit (%dC)", temp_severe);
	else if (enc_temp_current >= temp_warning)
	    sprintf(cp, " - over Warning Limit (%dC)", temp_warning);
    }

    *msg = msgbuf;
    return ecode;
}

int
check_ups(RTS *rp,
	  char **msg,
	  struct rts_vendor_parameters *vp,
	  struct rts_environmental_information *eip)
{
    static char msgbuf[256];
    int ecode = 0;


    *msgbuf = '\0';
    
#if RTS_VERSION >= 3
    switch (eip->ups_status)
    {
      case 0:
	break;
	
      case 1:
      case 6:
	sprintf(msgbuf, "UPS: Normal");
	break;
	
      case 7:
	sprintf(msgbuf, "UPS: AC Power Lost");
	ecode = 1;
	break;
	
      case 8:
	sprintf(msgbuf, "UPS: Two Minute Warning");
	ecode = 2;
	break;
	
      default:
	sprintf(msgbuf, "UPS: Unknown (#%d)", eip->ups_status);
	ecode = -1;
	break;
    }
#endif
    
    *msg = msgbuf;
    return ecode;
}


int
check_power(RTS *rp,
	    char **msg,
	    struct rts_vendor_parameters *vp,
	    struct rts_environmental_information *eip)
{
    static char msgbuf[256];
    static char msgbuf2[256];
    char *mp;
    int ecode = 0;
    int necode = 0;


    *msgbuf = '\0';
    mp = msgbuf2;
    
    necode = check_vt_status(eip->five_volt_status);
    if (necode > 0)
    {
	if (necode > ecode)
	    ecode = necode;
	
	sprintf(mp, "+5V: %.2f", rts_w2i(eip->five_volt_level) / 100.0);
    }

    necode = check_vt_status(eip->twelve_volt_status);
    if (necode > 0)
    {
	if (necode > ecode)
	    ecode = necode;
	
	if (*mp)
	{
	    while (*mp) ++mp;
	    *mp++ = ',';
	    *mp++ = ' ';
	}
	
	sprintf(mp, "+12V: %.2f", rts_w2i(eip->twelve_volt_level) / 100.0);
    }

    necode = check_vt_status(eip->scsi_termination_volt_status);
    if (necode > 0)
    {
	if (necode > ecode)
	    ecode = necode;
	
	if (*mp)
	{
	    while (*mp) ++mp;
	    *mp++ = ',';
	    *mp++ = ' ';
	}
	
	sprintf(mp, "SCSI: %.2f", rts_w2i(eip->scsi_termination_volt_level) / 100.0);
    }

    necode = check_vt_status(eip->dram_0_volt_status);
    if (necode > 0)
    {
	if (necode > ecode)
	    ecode = necode;
	
	if (*mp)
	{
	    while (*mp) ++mp;
	    *mp++ = ',';
	    *mp++ = ' ';
	}
	
	sprintf(mp, "DRAM 0: %.2f", rts_w2i(eip->dram_0_volt_level) / 100.0);
    }

    necode = check_vt_status(eip->dram_1_volt_status);
    if (necode > 0)
    {
	if (necode > ecode)
	    ecode = necode;

	if (*mp)
	{
	    while (*mp) ++mp;
	    *mp++ = ',';
	    *mp++ = ' ';
	}
	
	sprintf(mp, "DRAM 1: %.2f", rts_w2i(eip->dram_1_volt_level) / 100.0);
    }
    
    necode = check_vt_status(eip->backup_battery_volt_status);
    if (necode > 0)
    {
	if (necode > ecode)
	    ecode = necode;
	
	if (*mp)
	{
	    while (*mp) ++mp;
	    *mp++ = ',';
	    *mp++ = ' ';
	}
	
	sprintf(mp, "Battery: %.2f", rts_w2i(eip->backup_battery_volt_level) / 100.0);
    }

    switch (ecode)
    {
      case 0:
	sprintf(msgbuf, "Power: Normal");
	break;
	
      case 1:
	sprintf(msgbuf, "Power: Marginal (%s)", msgbuf2);
	break;

      case 2:
	sprintf(msgbuf, "Power: Abnormal (%s)", msgbuf2);
	break;
    }
	
    *msg = msgbuf;
    return ecode;
}


int
check_redgrps(RTS *rp,
	      char **msg)
{
    int i, code, ecode = 0;
    static char msgbuf[256];
    char *mp;
    
    int n_redgrps = 0;
    int n_degraded = 0;
    int n_offline = 0;
    int n_creating = 0;
    
    
    for (i = 0; i < 32; i++)
    {
	struct rts_redundancy_group_information rb;
	
	code = rts_read_redundancy_group_information(rp, i, &rb);
	if (code)
	{
	    rts_perror(stderr, "rts_read_redundancy_group_information",rp);
	    exit(-1);
	}

	if (rb.raid_status > 0)
	    n_redgrps++;
	
	switch (rb.raid_status)
	{
	  case 2:
	    n_degraded++;
	    break;

	  case 3:
	    n_offline++;
	    break;

	  case 4:
	    n_creating++;
	    break;
	}
    }

    sprintf(msgbuf, "RedGrps: %d", n_redgrps);
    mp = msgbuf;
    while (*mp) ++mp;

    if (n_offline > 0 || n_degraded > 0)
    {
	sprintf(mp, " (%d offline, %d degraded)", n_offline, n_degraded);
	if (n_offline > 0)
	    ecode = 2;
	else if (n_degraded > 0)
	    ecode = 1;
    }
    else
	sprintf(mp, " online");

    *msg = msgbuf;
    return ecode;
}


int
check_raidsets(RTS *rp,
	       char **msg)
{
    int i, code, ecode = 0;
    static char msgbuf[256];
    char *mp;
    
    int n_raidsets = 0;
    int n_degraded = 0;
    int n_offline = 0;
    int n_creating = 0;
    
    
    for (i = 0; i < 32; i++)
    {
	struct rts_raid_set_information rb;
	
	code = rts_read_raid_set_information(rp, i, &rb);
	if (code)
	{
	    rts_perror(stderr, "rts_read_raid_set_information",rp);
	    exit(-1);
	}

	if (rb.raid_status > 0)
	    n_raidsets++;
	
	switch (rb.raid_status)
	{
	  case 2:
	    n_degraded++;
	    break;

	  case 3:
	    n_offline++;
	    break;

	  case 4:
	    n_creating++;
	    break;
	}
    }

    sprintf(msgbuf, "RaidSets: %d", n_raidsets);
    mp = msgbuf;
    while (*mp) ++mp;

    if (n_offline > 0 || n_degraded > 0)
    {
	sprintf(mp, " - %d Offline & %d Degraded",
		n_offline, n_degraded);
	
	if (n_offline > 0)
	    ecode = 2;
	else if (n_degraded > 0)
	    ecode = 1;
    }
    else
	sprintf(mp, " Online");

    *msg = msgbuf;
    return ecode;
}


int
check_controllers(RTS *rp,
		  char **msg)
{
    int code, ecode = 0;
    static char msgbuf[256];
    char *mp;
    struct rts_redundant_controller_information rb;

    *msgbuf = '\0';
    
    code = rts_read_redundant_controller_information(rp, &rb);
    if (code)
    {
	rts_perror(stderr, "rts_read_redundant_controller_information", rp);
	exit(-1);
    }

    if ((rts_w2i(rb.flags) & 0x01) == 1)
    {
	/* Redundant controllers installed */

	mp = msgbuf;
	sprintf(mp, "RedCtrl: ");
	while (*mp) ++mp;
	
	switch (rb.status)
	{
	  case 0:
	    ecode = 1;
	    sprintf(mp, "Unbound");
	    break;

	  case 1:
	    ecode = 1;
	    sprintf(mp, "Attempting bind");
	    break;

	  case 2:
	    ecode = 2;
	    sprintf(mp, "Bind aborted ");
	    while (*mp) ++mp;
	    switch (rb.reason)
	    {
	      case 0:
		sprintf(mp, "(Timeout)");
		break;

	      case 1:
		sprintf(mp, "(Cache mismatch)");
		break;

	      case 2:
		sprintf(mp, "(Firmware version mismatch)");
		break;

	      case 3:
		sprintf(mp, "(Hardware mismatch)");
		break;

	      case 4:
		sprintf(mp, "(Hardware error)");
		break;

	      case 5:
		sprintf(mp, "(Parity error)");
		break;

	      case 6:
		sprintf(mp, "(Raid set error)");
		break;

	      case 7:
		sprintf(mp, "(Hostmap error)");
		break;

	      case 8:
		sprintf(mp, "(Unexpected reason)");
		break;

	      case 9:
		sprintf(mp, "(Bad response)");
		break;

	      case 10:
		sprintf(mp, "(Query received)");
		break;

	      case 11:
		sprintf(mp, "(Cache not mirrored)");
		break;

	      default:
		sprintf(mp, "(Reason #%d)", rb.reason);
	    }
	    break;

	  case 3:
	    ecode = 0;
	    sprintf(mp, "Normal");
	    break;

	  case 4:
	    ecode = 2;
	    sprintf(mp, "Controller set broken ");
	    while (*mp) ++mp;
	    switch (rb.reason)
	    {
	      case 0:
		sprintf(mp, "(Bugcheck)");
		break;

	      case 1:
		sprintf(mp, "(System restart command received)");
		break;

	      case 2:
		sprintf(mp, "(System shutdown command received)");
		break;

	      case 3:
		sprintf(mp, "(Raid set error)");
		break;

	      case 4:
		sprintf(mp, "(Query message received during bind)");
		break;

	      case 5:
		sprintf(mp, "(Communications Problems)");
		break;

	      default:
		sprintf(mp, "(Reason #%d)", rb.reason);
	    }
	    break;

	  case 5:
	    ecode = 2;
	    sprintf(mp, "Controller failed");
	    break;

	  case 6:
	    ecode = 1;
	    sprintf(mp, "Switchover in progress");
	    break;

	  case 7:
	  case 8:
	    if (rb.status == 7)
	    {
		ecode = 2;
		sprintf(mp, "Switchover failed");
	    }
	    else
	    {
		ecode = 1;
		sprintf(mp, "Switchover completed successfully");
	    }

	    while (*mp) ++mp;
	    switch (rb.reason)
	    {
	      case 0:
		sprintf(mp, "(Bugcheck)");
		break;

	      case 1:
		sprintf(mp, "(I/O hang)");
		break;

	      case 2:
		sprintf(mp, "(Missing heartbeat)");
		break;

	      case 3:
		sprintf(mp, "(Other controller failed to bind after power loss))");
		break;

	      case 4:
		sprintf(mp, "(Power was lost after other controller failed)");
		break;

	      default:
		sprintf(mp, "(Reason #%d)", rb.reason);
	    }
	    break;

	  default:
	    ecode = 1;
	    sprintf(mp, "Unknown status: %d", rb.status);
	}
    }

    *msg = msgbuf;
    return ecode;
}


int
main(int argc, char *argv[])
{
    RTS *rp;
    int c, code;
    time_t c_t;
    struct rts_vendor_parameters vpb;
    struct rts_environmental_information eib;
    struct rts_static_system_information ssib;
    struct rts_dynamic_system_information dsib;
    struct rts_system_parameters spb;
    char msgbuf[1024], *msg, *mp;

    
    argv0 = argv[0];

    /* If setuid - drop write privileges! */
    if (geteuid() != getuid())
	readonly = 1;
    
    while ((c = getopt(argc, argv, "VDHNARvesdyhrm")) != -1)
	switch (c)
	{
	  case 'V':
	    printf("Viper %s, built on %s %s, for ",
		   version, __DATE__, __TIME__);
	    printf("RTS Version: %d\n", RTS_VERSION);
	    exit(0);

	  case 'D':
	    debug = 1;
	    break;

	  case 'N':
	    check_netsaint = 1;
	    break;
	    
	  case 'H':
	    usage(stdout);
	    exit(0);

	  case 'A':
	    print_vendor = 1;
	    print_env = 1;
	    print_static = 1;
	    print_dynamic = 1;
	    print_system = 1;
	    print_host = 1;
	    print_redgrp = 1;
	    print_misc = 1;
	    break;

	  case 'R':
	    readonly = 1;
	    break;
	    
	  case 'v':
	    print_vendor = 1;
	    break;
	    
	  case 'e':
	    print_env = 1;
	    break;
	    
	  case 's':
	    print_static = 1;
	    break;
	    
	  case 'd':
	    print_dynamic = 1;
	    break;
	    
	  case 'y':
	    print_system = 1;
	    break;
	    
	  case 'h':
	    print_host = 1;
	    break;
	    
	  case 'r':
	    print_redgrp = 1;
	    break;
	    
	  case 'm':
	    print_misc = 1;
	    break;
	    
	  default:
	    usage(stderr);
	    exit(1);
	}
    
    if (optind >= argc)
    {
	usage(stderr);
	exit(1);
    }
    
    rp = rts_open(argv[optind]);
    
#if 0 /* Doesn't work - due to rts_lock()/rts_unlock() */
    /* If a setuid binary - drop all privs now that the device is opened */
    if (geteuid() != getuid())
    {
	setgid(getgid());
	setuid(getuid());
    }
#endif
    
    if (rp == NULL)
    {
	if (check_netsaint)
	{
	    fprintf(stdout,
		    "VIPER UNKNOWN: rts_open: %s: %s\n",
		    argv[optind], strerror(errno));
	    exit(-1);
	}

	fprintf(stderr, "rts_open: %s: %s\n", argv[optind], strerror(errno));
	exit(1);
    }

    if (!check_netsaint)
    {
	printf("[Viper %s, built on %s %s, for RTS version %d]\n",
	       version, __DATE__, __TIME__, RTS_VERSION);
#if 0
	printf("%-35s%s (%s)\n", 
	       "Controller Device:", argv[optind],
	       readonly ? "Read-Only" : "Read/Write");
	printf("%-35s%d\n", "Controller RTS version:",
	       rp->version);
#endif
    }
    
    ++optind;

    if (print_static)
    {
	puts("\n*** Static System Information:");
	
	code = rts_read_static_system_information(rp, &ssib);
	if (code)
	{
	    rts_perror(stderr, "rts_read_static_system_information", rp);
	    exit(1);
	}

	rts_print_static_system_information(stdout, RTS_VERSION, &ssib);
    }

    if (print_dynamic)
    {
	puts("\n*** Dynamic System Information:");
	
	code = rts_read_dynamic_system_information(rp, &dsib);
	if (code)
	{
	    rts_perror(stderr, "rts_read_dynamic_system_information", rp);
	    exit(1);
	}

	rts_print_dynamic_system_information(stdout, RTS_VERSION, &dsib);
    }

    if (print_vendor)
    {
	puts("\n*** Vendor Parameters:");
	
	code = rts_read_vendor_parameters(rp, &vpb);
	if (code != 0)
	{
	    rts_perror(stderr, "rts_read_vendor_parameters failed", rp);
	    exit(1);
	}

	rts_print_vendor_parameters(stdout, RTS_VERSION, &vpb);
    }

    if (print_env)
    {
	puts("\n*** Environmental Parameters:");
	
	code = rts_read_environmental_information(rp, &eib);
	if (code != 0)
	{
	    rts_perror(stderr, "rts_read_environmental_parameters", rp);
	    exit(1);
	}

	rts_print_environmental_information(stdout, RTS_VERSION, &eib);
    }


    if (print_system)
    {
	puts("\n*** System Parameters:");
	
	code = rts_read_system_parameters(rp, &spb);
	if (code)
	{
	    rts_perror(stderr, "rts_read_system_parameters", rp);
	    exit(1);
	}

	rts_print_system_parameters(stdout, RTS_VERSION, &spb);
    }

    if (print_host)
    {
	int i, flg;
	struct rts_host_parameters hpb;

	printf("\n*** Host Parameters:\n");
	flg = 0;
	for (i = 0; i < 4; i++)
	{
	    code = rts_read_host_parameters(rp, i, &hpb);
	    if (code)
	    {
		rts_perror(stderr, "rts_read_host_parameters", rp);
		exit(1);
	    }

	    if (flg)
		putchar('\n');
	    
	    rts_print_host_parameters(stdout, RTS_VERSION, &hpb);
	    flg = 1;
	}
    }

    if (print_redgrp)
    {
	int i;
	
	puts("\n*** Redundancy Group Information:");

	for (i = 0; i < 32; i++)
	{
	    struct rts_redundancy_group_information rb;

	    code = rts_read_redundancy_group_information(rp, i, &rb);
	    if (code)
	    {
		rts_perror(stderr, "rts_read_redundancy_group_information",rp);
		exit(1);
	    }

	    if (rb.raid_status != 0)
	    {
		rts_print_redundancy_group_information(stdout, RTS_VERSION, &rb);
		if (i+1 < 32)
		    putchar('\n');
	    }
	}
    }
    
    if (print_misc)
    {
	puts("\n*** Miscellaneous:");
	
	code = rts_read_date_and_time(rp, &c_t);
	if (code)
	{
	    rts_perror(stderr, "rts_read_date_and_time", rp);
	    exit(1);
	}
	
	printf("%-35s%s\n", "Clock on controller:", ctime(&c_t));
    }


    if (check_netsaint)
    {
	int ecode = 0;
	int necode = 0;

	
	code = rts_read_vendor_parameters(rp, &vpb);
	if (code != 0)
	{
	    rts_perror(stdout,
		       "VIPER UNKNOWN: rts_read_vendor_parameters failed",
		       rp);
	    exit(-1);
	}
	
	code = rts_read_environmental_information(rp, &eib);
	if (code != 0)
	{
	    rts_perror(stdout,
		       "VIPER UNKNOWN: rts_read_environmental_parameters",
		       rp);
	    exit(-1);
	}

	mp = msgbuf;
	
	ecode = 0;
	
	msg = NULL;
	necode = check_raidsets(rp, &msg);
	if (necode > ecode)
	    ecode = necode;

	if (msg && *msg)
	{
	    strcpy(mp, msg);
	    while (*mp) ++mp;
	}

	msg = NULL;
	necode = check_controllers(rp, &msg);
	if (necode > ecode)
	    ecode = necode;

	if (msg && *msg)
	{
	    *mp++ = ';';
	    *mp++ = ' ';
	    strcpy(mp, msg);
	    while (*mp) ++mp;
	}

	msg = NULL;
	necode = check_temp(rp, &msg, &vpb, &eib);
	if (necode > ecode)
	    ecode = necode;

	if (msg && *msg)
	{
	    *mp++ = ';';
	    *mp++ = ' ';
	    strcpy(mp, msg);
	    while (*mp) ++mp;
	}
	    
	msg = NULL;
	necode = check_power(rp, &msg, &vpb, &eib);
	if (necode > ecode)
	    ecode = necode;

	if (msg && *msg)
	{
	    *mp++ = ';';
	    *mp++ = ' ';
	    strcpy(mp, msg);
	    while (*mp) ++mp;
	}

	msg = NULL;
	necode = check_ups(rp, &msg, &vpb, &eib);
	if (necode > ecode)
	    ecode = necode;
	
	if (msg && *msg)
	{
	    *mp++ = ';';
	    *mp++ = ' ';
	    strcpy(mp, msg);
	    while (*mp) ++mp;
	}

	switch (ecode)
	{
	  case 0:
	    printf("VIPER OK: %s\n", msgbuf);
	    break;
	    
	  case 1:
	    printf("VIPER WARNING: %s\n", msgbuf);
	    break;
	    
	  case 2:
	    printf("VIPER CRITICAL: %s\n", msgbuf);
	    break;
	    
	  default:
	    printf("VIPER UNKNOWN: code=%d, errno(%d)=%s\n",
		   ecode,
		   errno,
		   strerror(errno));
	    exit(-1);
	}

	exit(ecode);
    }
    
    exit(0);
}
